/* 3d.c                         */
/* 3d base functions of the poor*/
/* Part of libcaca demo */

#include <demo.h>
#include <3d.h>
#include <sprites.h>

/* Must be called before any attempt
   to use 3d engine */
int init3dEngine(void)
{
  scene = malloc(sizeof(sScene));

  if(scene==NULL)
    {
      fprintf(stderr, "Out of memory\n");
      return 0;
    }

  scene->objectCount = 0;
  scene->backgroundColor.r = 0;
  scene->backgroundColor.g = 0;
  scene->backgroundColor.b = 0;

  scene->object = malloc(1);


  scene->camera.position.x=0;
  scene->camera.position.y=0;
  scene->camera.position.z=0;
  scene->camera.rotation.x=0;
  scene->camera.rotation.y=0;
  scene->camera.rotation.z=0;
  scene->camera.pivot.x=0;
  scene->camera.pivot.y=0;
  scene->camera.pivot.z=0;



  scene->zbuffer = malloc(sizeof(int)*XSIZ*YSIZ);
  memset(scene->zbuffer, 254, XSIZ*YSIZ*4);
  return 1;
}

/* Must be called if you want noone
   to bore you while reading sources */
void close3dEngine(void)
{
  int i;
  int f;
  if(scene)
    {
      if(scene->objectCount)
	{
	  for(i=0;i<scene->objectCount;i++)
	    {
	      if(scene->object[i].vertex)
		free(scene->object[i].vertex);
	      if(scene->object[i].rvertex)
		free(scene->object[i].rvertex);
	      if(scene->object[i].tvertex)
		free(scene->object[i].tvertex);
	      if(scene->object[i].d2vertex)
		free(scene->object[i].d2vertex);
	      

	      if(scene->object[i].face)
		{
		  for(f=0;f<scene->object[i].faceCount;f++)
		    {
		      free(scene->object[i].face[f].normal);
		    }
		  free(scene->object[i].face);
		}
	    }
	  if(scene->object)
	    free(scene->object);

	}
      if(scene->zbuffer)
	free(scene->zbuffer);

      free(scene);
    } 
}

/* Object properties */
void moveObject(int id, float x, float y, float z)
{
  scene->object[id].position.x = x;
  scene->object[id].position.y = y;
  scene->object[id].position.z = z;
}
void rotateObject(int id, float x, float y, float z)
{
  scene->object[id].rotation.x = DEG2RAD(x);
  scene->object[id].rotation.y = DEG2RAD(y);
  scene->object[id].rotation.z = DEG2RAD(z);

}
void pivotObject(int id, float x, float y, float z)
{
  scene->object[id].pivot.x = x;
  scene->object[id].pivot.y = y;
  scene->object[id].pivot.z = z;
}
void moveCamera(float x, float y, float z)
{
  scene->camera.position.x = x;
  scene->camera.position.y = y;
  scene->camera.position.z = z;
}
void rotateCamera(float x, float y, float z)
{
  scene->camera.rotation.x = DEG2RAD(x);
  scene->camera.rotation.y = DEG2RAD(y);
  scene->camera.rotation.z = DEG2RAD(z);

}
void pivotCamera(float x, float y, float z)
{
  scene->camera.pivot.x = x;
  scene->camera.pivot.y = y;
  scene->camera.pivot.z = z;
}



void colorObject(int id, int r, int g, int b)
{
  int i;
  for(i=0;i<scene->object[id].faceCount;i++)
    {
      scene->object[id].face[i].color.r = r;
      scene->object[id].face[i].color.g = g;
      scene->object[id].face[i].color.b = b;
    }
}
void objectRenderMode(int id, int mode)
{
  scene->object[id].renderMode = mode;
}
void objectEnableCulling(int id)
{
  scene->object[id].culling = 1;
}
void objectDisableCulling(int id)
{
  scene->object[id].culling = 0;
}
void objectSetVisible(int id)
{
  scene->object[id].visible = 1;
}
void objectSetInvisible(int id)
{
  scene->object[id].visible = 0;
}

sObject *objectGet(int id)
{
  return &scene->object[id];
}


/* Object management */

int addObject(int vtxCount, int faceCount)
{
  int i;
  scene->object = realloc(scene->object, 
			  sizeof(sObject) * (scene->objectCount+1));

  scene->object[scene->objectCount].vertexCount = vtxCount;
  scene->object[scene->objectCount].vertex = malloc(sizeof(sVector)*vtxCount);
  scene->object[scene->objectCount].tvertex = malloc(sizeof(sVector)*vtxCount);
  scene->object[scene->objectCount].rvertex = malloc(sizeof(sVector)*vtxCount);
  scene->object[scene->objectCount].d2vertex = malloc(sizeof(sVector)*vtxCount);

  scene->object[scene->objectCount].faceCount = faceCount;
  scene->object[scene->objectCount].face = malloc(sizeof(sFace)*faceCount);
  for(i=0;i<faceCount;i++)
    scene->object[scene->objectCount].face[i].normal = malloc(sizeof(sVector));
  

  scene->object[scene->objectCount].position.x = 0;
  scene->object[scene->objectCount].position.y = 0;
  scene->object[scene->objectCount].position.z = 0;
  scene->object[scene->objectCount].rotation.x = 0;
  scene->object[scene->objectCount].rotation.y = 0;
  scene->object[scene->objectCount].rotation.z = 0;
  scene->object[scene->objectCount].pivot.x = 0;
  scene->object[scene->objectCount].pivot.y = 0;
  scene->object[scene->objectCount].pivot.z = 0;

  scene->object[scene->objectCount].renderMode = RENDER_MODE_FLAT_Z;
  
  scene->object[scene->objectCount].visible = 1;

  scene->object[scene->objectCount].culling = 1;


  scene->objectCount++;

  return scene->objectCount-1;
}
/* Stolen from Optimum's galaxy effect 
   Writing this part is quite time consuming
   and not very interesting. And Cortex gave
   it to the world. Finnaly he ruled d4 w0rld.
*/

int addGalaxy(int starsCount, float radius)
{
  int id, i;
 
  id = addObject(starsCount, 1);

  for (i=0; i<starsCount; i++)
    {
      float r,th,h,dth;
      r = rand()*1.0 / RAND_MAX;
      r = (1-r)*(1-r)+0.05;
      if (r<0.12)
	th = rand()*M_PI*2/RAND_MAX;
      else 
	{
	  th = (rand()&3)*M_PI / 2+ r*r*2;
	  dth =   rand()*1.0/RAND_MAX;
	  dth = dth*dth*2;
	  th+=dth;
	}
      scene->object[id].vertex[i].y = radius*r*cos(th);
      scene->object[id].vertex[i].z = radius*r*sin(th);
      h = (1+cos(r*M_PI))*150;
      dth =   rand()*1.0/RAND_MAX;
      scene->object[id].vertex[i].x =h*(dth-0.5);
      scene->object[id].vertex[i].r = (2-r)*60+31;
    }

  return id;
}

int addCube(float width, float height, float depth, float x, float y, float z)
{
  int id;

  id = addObject(8, 12);


  scene->object[id].vertex[0].x = -(width/2) + x;
  scene->object[id].vertex[0].y = -(height/2)+ y;
  scene->object[id].vertex[0].z = -(depth/2) + z;

  scene->object[id].vertex[1].x = (width/2) + x;
  scene->object[id].vertex[1].y = -(height/2)+ y;
  scene->object[id].vertex[1].z = -(depth/2) + z;

  scene->object[id].vertex[2].x = (width/2) + x;
  scene->object[id].vertex[2].y = (height/2)+ y;
  scene->object[id].vertex[2].z = -(depth/2) + z;

  scene->object[id].vertex[3].x = -(width/2) + x;
  scene->object[id].vertex[3].y = (height/2)+ y;
  scene->object[id].vertex[3].z = -(depth/2) + z;

  scene->object[id].vertex[4].x = -(width/2) + x;
  scene->object[id].vertex[4].y = -(height/2)+ y;
  scene->object[id].vertex[4].z = (depth/2) + z;

  scene->object[id].vertex[5].x = (width/2) + x;
  scene->object[id].vertex[5].y = -(height/2)+ y;
  scene->object[id].vertex[5].z = (depth/2) + z;

  scene->object[id].vertex[6].x = (width/2) + x;
  scene->object[id].vertex[6].y = (height/2)+ y;
  scene->object[id].vertex[6].z = (depth/2) + z;

  scene->object[id].vertex[7].x = -(width/2) + x;
  scene->object[id].vertex[7].y = (height/2)+ y;
  scene->object[id].vertex[7].z = (depth/2) + z;


  scene->object[id].face[0].a = 0; /*Front*/
  scene->object[id].face[0].b = 1;
  scene->object[id].face[0].c = 2;
  scene->object[id].face[1].a = 0;
  scene->object[id].face[1].b = 2;
  scene->object[id].face[1].c = 3;
 
  scene->object[id].face[2].a = 4; /*Top*/
  scene->object[id].face[2].b = 5;
  scene->object[id].face[2].c = 1;
  scene->object[id].face[3].a = 4;
  scene->object[id].face[3].b = 1;
  scene->object[id].face[3].c = 0;

  scene->object[id].face[4].a = 1; /*Right*/
  scene->object[id].face[4].b = 5;
  scene->object[id].face[4].c = 6;
  scene->object[id].face[5].a = 1;
  scene->object[id].face[5].b = 6;
  scene->object[id].face[5].c = 2;

  scene->object[id].face[6].a = 5; /*Far*/
  scene->object[id].face[6].b = 4;
  scene->object[id].face[6].c = 7;
  scene->object[id].face[7].a = 5;
  scene->object[id].face[7].b = 7;
  scene->object[id].face[7].c = 6;

  scene->object[id].face[8].a = 4; /*Left*/
  scene->object[id].face[8].b = 0;
  scene->object[id].face[8].c = 3;
  scene->object[id].face[9].a = 4;
  scene->object[id].face[9].b = 3;
  scene->object[id].face[9].c = 7;

  scene->object[id].face[10].a = 6; /*Bottom*/
  scene->object[id].face[10].b = 7;
  scene->object[id].face[10].c = 3;
  scene->object[id].face[11].a = 6;
  scene->object[id].face[11].b = 3;
  scene->object[id].face[11].c = 2;
  
  objectNormals(id);

  return id;
} 


int addTriObject(unsigned char *name)
{
  FILE *fp;
  int triCount=0;
  unsigned char line[1024];
  int id;
  float x1,y1,z1;
  float x2,y2,z2;
  float x3,y3,z3;
  int color;
  int vtx=0;
  int face=0;
  fp = fopen(name, "r+");
  if(fp==NULL)
    {
      fprintf(stderr,"Can't open %s for reading\n", name);
      return 0;
    }
  while(!feof(fp))
    {
      fgets(line, 1024, fp);
      triCount++;
    }

  id = addObject(triCount*3,triCount);
 rewind(fp);

  while(!feof(fp))
    {
     fgets(line, 1024, fp);
     sscanf(line, "%f %f %f %f %f %f %f %f %f 0x%X\n", 
	    &x1,&y1,&z1,
	    &x2,&y2,&z2,
	    &x3,&y3,&z3,
	    &color);

     scene->object[id].vertex[vtx].x = x1*SCALE_FACTOR;
     scene->object[id].vertex[vtx].y = y1*SCALE_FACTOR;
     scene->object[id].vertex[vtx].z = z1*SCALE_FACTOR;
     scene->object[id].vertex[vtx+1].x = x2*SCALE_FACTOR;
     scene->object[id].vertex[vtx+1].y = y2*SCALE_FACTOR;
     scene->object[id].vertex[vtx+1].z = z2*SCALE_FACTOR;
     scene->object[id].vertex[vtx+2].x = x3*SCALE_FACTOR;
     scene->object[id].vertex[vtx+2].y = y3*SCALE_FACTOR;
     scene->object[id].vertex[vtx+2].z = z3*SCALE_FACTOR;

     scene->object[id].face[face].c = vtx;
     scene->object[id].face[face].b = vtx+1;
     scene->object[id].face[face].a = vtx+2;

     scene->object[id].face[face].color.r = (color&0xFF0000)>>16;
     scene->object[id].face[face].color.g = (color&0x00FF00)>>8;
     scene->object[id].face[face].color.b = (color&0x0000FF);


     
     face++;
     vtx+=3;


    }


 fclose(fp); 
 return id;

}



/* Object Transformations */
void transformObjects(void)
{
  int id, v;
  for(id=0;id<scene->objectCount;id++)
    {
 	  for(v=0;v<scene->object[id].vertexCount;v++)
	    {
	      /* Object space */
	      rotateVertex(&scene->object[id].rvertex[v], 
			   &scene->object[id].vertex[v], 
			   &scene->object[id].rotation, 
			   &scene->object[id].pivot);
	      translateVertex(&scene->object[id].tvertex[v], 
			      &scene->object[id].rvertex[v], 
			      &scene->object[id].position);

	      /* World space*/	      
	      rotateVertex(&scene->object[id].rvertex[v], 
			   &scene->object[id].tvertex[v], 
			   &scene->camera.rotation, 
			   &scene->camera.pivot);
	      translateVertex(&scene->object[id].tvertex[v], 
			      &scene->object[id].rvertex[v], 
			      &scene->camera.position);


	      projectVertex(&scene->object[id].d2vertex[v], 
			    &scene->object[id].tvertex[v]);
	      
	    }
	 
	  if(scene->object[id].faceCount)
	    objectNormals(id);
    }


}

void normalVertex(int objid, int faceid)
{
  float x1 = scene->object[objid].rvertex[scene->object[objid].face[faceid].a].x;
  float y1 = scene->object[objid].rvertex[scene->object[objid].face[faceid].a].y;
  float z1 = scene->object[objid].rvertex[scene->object[objid].face[faceid].a].z;
  float x2 = scene->object[objid].rvertex[scene->object[objid].face[faceid].b].x;
  float y2 = scene->object[objid].rvertex[scene->object[objid].face[faceid].b].y;
  float z2 = scene->object[objid].rvertex[scene->object[objid].face[faceid].b].z;
  float x3 = scene->object[objid].rvertex[scene->object[objid].face[faceid].c].x;
  float y3 = scene->object[objid].rvertex[scene->object[objid].face[faceid].c].y;
  float z3 = scene->object[objid].rvertex[scene->object[objid].face[faceid].c].z;

  scene->object[objid].face[faceid].normal->x = ((y1-y2)*(z1-z3))-((z1-z2)*(y1-y3));
  scene->object[objid].face[faceid].normal->y = ((z1-z2)*(x1-x3))-((x1-x2)*(z1-z3));
  scene->object[objid].face[faceid].normal->z = ((x1-x2)*(y1-y3))-((y1-y2)*(x1-x3));

  normalizeVector(scene->object[objid].face[faceid].normal);
  /*
  if(faceid == 0)
    printf("Normal of face 0 : %f %f %f\n", scene->object[objid].face[faceid].normal.x, scene->object[objid].face[faceid].normal.y, scene->object[objid].face[faceid].normal.z);
  */

}

void objectNormals(int objid)
{
  int faces;
  for(faces=0;faces<scene->object[objid].faceCount;faces++)
    {
      normalVertex(objid,faces);
    }
}

void lightObject(int objid, sVector *light)
{
  int face;
 

  for(face=0;face<scene->object[objid].faceCount;face++)
    {
      sVector *a = scene->object[objid].face[face].normal;
      sVector *b = light;
      
      scene->object[objid].face[face].l  = (a->x*b->x)+(a->y*b->y)+(a->z*b->z);

      scene->object[objid].face[face].l+=1.5f;
      scene->object[objid].face[face].l/=3.0f;

    }


}


void normalizeVector(sVector *vect)
{
  int nlength = caca_sqrt((vect->x*vect->x)+(vect->y*vect->y)+(vect->z*vect->z));
  vect->x = vect->x / (nlength?nlength:1);
  vect->y = vect->y / (nlength?nlength:1);
  vect->z = vect->z / (nlength?nlength:1);
}

void rotateVertex(sVector *tvertex, sVector *vertex, sVector *rotation, sVector *pivot)
{
  float x,y,z;
  float sinx,siny,sinz;
  float cosx,cosy,cosz;

  tvertex->x = vertex->x;
  tvertex->y = vertex->y;
  tvertex->z = vertex->z;

  if(rotation->x!=0)
    {
      sinx = sin(rotation->x);
      cosx = cos(rotation->x);
      
      y = pivot->y+(tvertex->y * cosx) - (tvertex->z * sinx);
      z = pivot->z+(tvertex->z * cosx) + (tvertex->y * sinx);
      tvertex->y=y;
      tvertex->z=z;

    }
  if(rotation->y!=0)
    {

      siny = sin(rotation->y);
      cosy = cos(rotation->y);

      z = pivot->z+(tvertex->z * cosy) - (tvertex->x * siny);
      x = pivot->x+(tvertex->x * cosy) + (tvertex->z * siny);
      tvertex->z=z;
      tvertex->x=x;

    }
  if(rotation->z!=0)
    {
      sinz = sin(rotation->z);
      cosz = cos(rotation->z);

      x = pivot->x+(tvertex->x * cosz) - (tvertex->y * sinz);
      y = pivot->y+(tvertex->y * cosz) + (tvertex->x * sinz);
      tvertex->x=x;
      tvertex->y=y;

    }      
}
void translateVertex(sVector *tvertex, sVector *vertex, sVector *translation)
{
  tvertex->x=vertex->x+translation->x;
  tvertex->y=vertex->y+translation->y;
  tvertex->z=vertex->z+translation->z;
}

void projectVertex(sVector *d2vertex, sVector *tvertex)
{
  d2vertex->z = tvertex->z;
  d2vertex->x = (320*tvertex->x/tvertex->z)+XSIZ/2;
  d2vertex->y = (320*tvertex->y/tvertex->z)+YSIZ/2;
}

/* Object display */

void drawObjects(void)
{
  int id;
  for(id=0;id<scene->objectCount;id++)
    {
      if(scene->object[id].visible)
	switch(scene->object[id].renderMode)
	  {
	case RENDER_MODE_FLAT:
	  drawObjectFlat(id);
	  break;
	case RENDER_MODE_FLAT_Z:
	  drawObjectFlatZbuffer(id);
	  break;
	case RENDER_MODE_DOT:
	  drawObjectDot(id);
	  break;
	default:
	  drawObjectFlat(id);
	  break;
	  
	}
    }
}


void drawObjectFlatZbuffer(int id)
{
  int face;
  sVector light;
  sColor col;
 
  light.x = 1.0;
  light.y = 1.0;
  light.z = 0.5;


  lightObject(id, &light);
  for(face=0;face<scene->object[id].faceCount;face++)
    {
      sVector *A = &scene->object[id].d2vertex[scene->object[id].face[face].a];
      sVector *B = &scene->object[id].d2vertex[scene->object[id].face[face].b];
      sVector *C = &scene->object[id].d2vertex[scene->object[id].face[face].c];

      if(!scene->object[id].culling)
	goto noculling;
      /* 2d Backface Culling */
      if (((B->y-A->y)/(B->x-A->x) - (C->y-A->y)/(C->x-A->x) < 0) ^ (A->x <= B->x == A->x > C->x))
	{
	noculling:
	 
	  col.r = scene->object[id].face[face].color.r * scene->object[id].face[face].l;
	  col.g = scene->object[id].face[face].color.g * scene->object[id].face[face].l;
	  col.b = scene->object[id].face[face].color.b * scene->object[id].face[face].l;
	  
	  
	  drawTriangleFlatZBuffer((int*)screen, 
				  &scene->object[id].d2vertex[scene->object[id].face[face].a],
				  &scene->object[id].d2vertex[scene->object[id].face[face].b],
				  &scene->object[id].d2vertex[scene->object[id].face[face].c],
				  &col);
	 



	}
    }
}
void drawObjectFlat(int id)
{
  int face;
  sVector light;

  light.x = 1.0;
  light.y = 1.0;
  light.z = 0.5;


  lightObject(id, &light);

  for(face=0;face<scene->object[id].faceCount;face++)
    {
      sVector *A = &scene->object[id].d2vertex[scene->object[id].face[face].a];
      sVector *B = &scene->object[id].d2vertex[scene->object[id].face[face].b];
      sVector *C = &scene->object[id].d2vertex[scene->object[id].face[face].c];

      
      /* 2d Backface Culling */
      if (((B->y-A->y)/(B->x-A->x)-(C->y-A->y)/(C->x-A->x)<0)^(A->x<=B->x==A->x>C->x))
	{
	  sColor col;
	  col.r = scene->object[id].face[face].color.r * scene->object[id].face[face].l;
	  col.g = scene->object[id].face[face].color.g * scene->object[id].face[face].l;
	  col.b = scene->object[id].face[face].color.b * scene->object[id].face[face].l;


	  drawTriangleFlat((int*)screen, 
				&scene->object[id].d2vertex[scene->object[id].face[face].a],
				&scene->object[id].d2vertex[scene->object[id].face[face].b],
				&scene->object[id].d2vertex[scene->object[id].face[face].c],
				&col);
	
      }
    }
}

void drawObjectDot(int id)
{
  int face;
  float z,zc;
  float zmin=0, zmax=0;
  for(face=0;face<scene->object[id].vertexCount;face++)
    {
      if((scene->object[id].d2vertex[face].x < XSIZ) &&
	 (scene->object[id].d2vertex[face].x > 0) &&
	 (scene->object[id].d2vertex[face].y < YSIZ) &&
	 (scene->object[id].d2vertex[face].y > 0))
	{
	  z=scene->object[id].d2vertex[face].z;
	  zc = (scene->object[id].d2vertex[face].z-scene->object[id].position.z);

	 
	  zc = 255.0f-((scene->object[id].d2vertex[face].z-scene->object[id].position.z) + 450)/(900.0f/255.0f);




	  putSquareAdditive((int*)screentemp, 
		    scene->object[id].d2vertex[face].x, 
		    scene->object[id].d2vertex[face].y, 
		    15-z/200,
		    ((unsigned char)(zc*0.8f)),((unsigned char)(zc*0.5f)),((unsigned char)zc));
	

	  if(zc<zmin)
	    zmin=zc;
	  if(zc>zmax)
	    zmax=zc;
	}
    }
 
}



/* Primitives Drawing*/

/*Stolen from Sam's libcaca caca_fill_triangle*/
/*I'm a bit bored rewriting trifillers*/

void drawTriangleFlat(int *buffer, sVector *vertex1, sVector *vertex2, sVector *vertex3, sColor *color)
{
    int y, xa, xb, xmax, ymax;
    int x1 = vertex1->x, yp1 = vertex1->y;
    int x2 = vertex2->x, y2 = vertex2->y;
    int x3 = vertex3->x, y3 = vertex3->y;
    int col;

    /* Bubble-sort yp1 <= y2 <= y3 */
    if(yp1 > y2)
    {
        drawTriangleFlat(buffer,vertex2,vertex1,vertex3, color);
        return;
    }

    if(y2 > y3)
    {
        drawTriangleFlat(buffer,vertex1,vertex3,vertex2, color);
        return;
    }

    col = MAKE_COLOR(color);
    /* Promote precision */
    x1 *= 4;
    x2 *= 4;
    x3 *= 4;

    xmax = XSIZ - 1;
    ymax = YSIZ - 1;

    /* Rasterize our triangle */
    for(y = yp1 < 0 ? 0 : yp1; y <= y3 && y <= ymax; y++)
    {
        if(y <= y2)
        {
            xa = (yp1 == y2) ? x2 : x1 + (x2 - x1) * (y - yp1) / (y2 - yp1);
            xb = (yp1 == y3) ? x3 : x1 + (x3 - x1) * (y - yp1) / (y3 - yp1);
        }
        else
        {
            xa = (y3 == y2) ? x2 : x3 + (x2 - x3) * (y - y3) / (y2 - y3);
            xb = (y3 == yp1) ? x1 : x3 + (x1 - x3) * (y - y3) / (yp1 - y3);
        }

        if(xb < xa)
        {
            int tmp = xb;
            xb = xa; xa = tmp;
        }

        /* Rescale xa and xb, slightly cropping */
        xa = (xa + 2) / 4;
        xb = (xb - 2) / 4;

        if(xb < 0) continue;
        if(xa > xmax) continue;
        if(xa < 0) xa = 0;
        if(xb > xmax) xb = xmax;
	scanlineFlat(buffer, xa, xb, y, col);
        
    }
}
void scanlineFlat(int *buffer, int x1, int x2, int y, int color)
{
  int x;
  for(x = x1; x <= x2; x++)
    buffer[x+y*XSIZ] = color;
}





void drawTriangleFlatZBuffer(int *buffer, sVector *vertex1, sVector *vertex2, sVector *vertex3, sColor *color)
{
    int y, xa, xb, xmax, ymax, za, zb;
    int x1 = vertex1->x, yt1 = vertex1->y, z1 = (vertex1->z)*65535;
    int x2 = vertex2->x, y2 = vertex2->y, z2 = (vertex2->z)*65535;
    int x3 = vertex3->x, y3 = vertex3->y, z3 = (vertex3->z)*65535;
    int col;
    int offsety1, offsety2, offsety3;

    /* Bubble-sort yt1 <= y2 <= y3 */
    if(yt1 > y2)
    {
        drawTriangleFlatZBuffer(buffer,vertex2,vertex1,vertex3, color);
        return;
    }

    if(y2 > y3)
    {
        drawTriangleFlatZBuffer(buffer,vertex1,vertex3,vertex2, color);
        return;
    }


    col = MAKE_COLOR(color);
    /* Promote precision */
    /* And in fact we don't care */
    /*    x1 *= 4;
    x2 *= 4;
    x3 *= 4;
    */
    xmax = XSIZ - 1;
    ymax = YSIZ - 1;

    /* Rasterize our triangle */
    for(y = yt1 < 0 ? 0 : yt1; y <= y3 && y <= ymax; y++)
      {
        if(y <= y2)
	  {
	    offsety1=  (y2 - yt1);
	    offsety2 = (y3 - yt1);
	    offsety3 = (y - yt1);

            xa = (yt1 == y2) ? x2 : x1 + (((x2 - x1) * offsety3) / offsety1);
            xb = (yt1 == y3) ? x3 : x1 + (((x3 - x1) * offsety3) / offsety2);
            za = (yt1 == y2) ? z2 : z1 + (((z2 - z1) * offsety3) / offsety1);
            zb = (yt1 == y3) ? z3 : z1 + (((z3 - z1) * offsety3) / offsety2);
	  }
        else
	  {
	    offsety1 = (y - y3);
	    offsety2 = (y2 - y3);
	    offsety3 = (yt1 - y3);

            xa = (y3 == y2)  ? x2 : x3 +  ((x2 - x3)  * offsety1  / offsety2);
            xb = (y3 == yt1) ? x1 : x3 +  ((x1 - x3)  * offsety1  / offsety3);
            za = (y3 == y2)  ? z2 : z3 +  ((z2 - z3)  * offsety1  / offsety2);
            zb = (y3 == yt1) ? z1 : z3 +  ((z1 - z3)  * offsety1  / offsety3);
	  }

        if(xb < xa)
	  {
            int tmp = xb;
            xb = xa; xa = tmp;
	    tmp = zb;
	    zb = za; za = tmp;
        }

	/*    xa = (xa) / 4;
        xb = (xb) / 4;
        za = (za) / 4;
        zb = (zb) / 4;
	*/

        if(xb < 0) continue;
        if(xa > xmax) continue;
        if(xa < 0) xa = 0;
        if(xb > xmax) xb = xmax;

	scanlineFlatZBuffer(buffer, xa, xb, za, zb,y, col);
        
    }
}


inline void scanlineFlatZBuffer(int *buffer, int x1, int x2, int z1, int z2, int y, int color)
{
  int x;
  signed int z=z1;
  signed int zi = ((1.0f*z1)/((1.0f*z2))-1);
  int offset=y*XSIZ;
  
  for(x = x1; x <= x2; x++)
    {
      if((z)<((signed int)scene->zbuffer[x+offset]) && (z>0))
	{
	  buffer[x+offset] = color;
	  scene->zbuffer[x+offset] = z;
	}
      z+=zi;
    }
}



/*Postprocessing*/
#define CARTOON_SIZE 2
#define CARTOON_LEVEL 600000
void Cartoon(unsigned int *buffer)
{
  int i,j;
  int h,v;

  for(i=XSIZ;i<(XSIZ*(YSIZ-1));i++)
    {
      if(scene->zbuffer[i]<scene->zbuffer[i+1])
	h = scene->zbuffer[i+1] - scene->zbuffer[i];
      else
	h = scene->zbuffer[i] - scene->zbuffer[i+1];


      if(scene->zbuffer[i]<scene->zbuffer[i+XSIZ])
	v = scene->zbuffer[i+XSIZ] - scene->zbuffer[i];
      else
	v = scene->zbuffer[i] - scene->zbuffer[i+XSIZ];



      if((h > CARTOON_LEVEL) || (v > CARTOON_LEVEL))
	{
	  for(j=0;j<CARTOON_SIZE;j++)
	    buffer[i+(j-CARTOON_SIZE)] = 0x00FFFFFF;
	}

    }

}
