#include <demo.h>
#include <3d.h>
#include <font.h>
#include <interpolation.h>
#include <metaballs.h>
#include <bump.h>
#include <blur.h>
#include <tunnel.h>
#include <particles.h>
#include <raytracer.h>
#include <fire.h>

void renderBuffer(void);
void renderScreen(void);


void dumpPPM(char *filename);
int dump=0;
void displayPart2(int ticks);
void displayPart3(int ticks);

MODULE *module;
int musicPlay = 0;

struct caca_bitmap *bitmap;

#ifdef __LINUX__
pthread_t demoThread, musicThread;
#endif

struct timeval tv;
struct timezone tz;
long long baseTime = -1;
int frame=1; unsigned char fpsdisp[1024];
int curTime=0, oldTime=0;
int i=0;

/* ************** */
/* Demo variables */
/*                */
/* ************** */
int curPart=0;
float x,y,z;
/* Part 1*/
int mandaObject;
int mandaParticles;
int mandaParticles2;
sCurve *mandaCurve;
int medParticles;
sCurve *medCurve;
sCurve *part2Curve;

/* Part 2*/
#define PART2_COUNT 6
int part2Particles[PART2_COUNT];
float part2Time[PART2_COUNT];

sSprite *titlesprite;
#define BALL_COUNT 4
int balls;
int etoileObject;

sOutline *cabaret;

int main(int argc, char *argv[])
{
  int p;
  gettimeofday(&tv, &tz);
  
  curTime = (((tv.tv_sec*1000000) + tv.tv_usec));




  /* libcaca inits */
  if(caca_init() < 0)
    return 1;


  bitmap = caca_create_bitmap(32, XSIZ, YSIZ, XSIZ*4, 0xff0000, 0x00ff00, 0x0000ff, 0);


  /* no display if freetype inited after libcaca */
  if(!initFont())
    {
      fprintf(stderr, "Can initialize fonts.\n");
      return -1;
    }


  MikMod_RegisterAllDrivers();
  MikMod_RegisterAllLoaders();

  if (MikMod_Init("")) 
    {
      fprintf(stderr, "Could not initialize sound, reason: %s\n",
	      MikMod_strerror(MikMod_errno));
      return 1;
    }
  
  module = Player_Load("data/med_Funkytra.xm", 64, 0);
  if (!module) 
    {
      fprintf(stderr, "Could not load module, reason: %s\n",
	      MikMod_strerror(MikMod_errno));
      
    } 
  if(!init3dEngine())
    {
      fprintf(stderr, "Can initialize 3d engine.\n");
      return -1;
    }
  gettimeofday(&tv, &tz);
  baseTime = (tv.tv_sec*1000000) + tv.tv_usec;

 


 

  /* Demo init */
  titlesprite = loadPng("data/title.png");
  if(titlesprite==NULL)
    printf("Failed to load png\n");
  initBump();
  initRadialBlur();
 
  mandaObject = addTriObject("data/manda.tri");
  mandaCurve = loadCurve("data/curve.ac");
  mandaParticles = initParticles(150, 20, 0.1,0.001,-1.0, 0.5, "data/particle.png");
  mandaParticles2 = initParticles(150, 50, 0.1,0.001,-1.0, 0.5, "data/particle.png");


  cabaret = loadOutline("data/cabaret.ac");
  

  for(i=0;i<PART2_COUNT;i++)
    {
      p = caca_rand(0,3);
      switch(p)
	{
	case 0:
	  part2Particles[i] = initParticles(50, 20, 0.01,0.00001,-1.0, 0.05, "data/particle_r.png");
	  break;
	case 1:
	  part2Particles[i] = initParticles(50, 20, 0.01,0.00001,-1.0, 0.05, "data/particle_g.png");
	  break;
	case 2:
	  part2Particles[i] = initParticles(50, 20, 0.01,0.00001,-1.0, 0.05, "data/particle_b.png");
	  break;
	default:
	  part2Particles[i] = initParticles(50, 20, 0.01,0.00001,-1.0, 0.05, "data/particle_w.png");
	  break;

	}

      part2Time[i] = (1.0f/PART2_COUNT)*i*1.0f;


    }
  part2Curve = loadCurve("data/part1.ac");
  /*int count, float lifetime, float max_speed, float gravity, float up_speed, float far*/

  balls =  initMetaballs(BALL_COUNT);
 for(p=0;p<BALL_COUNT;p++)
    {
      scene->object[balls].vertex[p].color.r = caca_rand(0,1)*255;
      scene->object[balls].vertex[p].color.g = caca_rand(0,0)*255;
      scene->object[balls].vertex[p].color.b = caca_rand(0,0)*255;
    }

 etoileObject = addTriObject("data/etoile.tri");

  musicPlay = 1;
  pthread_create( &musicThread, NULL, musicFunction , (void*) NULL);
  pthread_create( &demoThread, NULL, demoFunction , (void*) NULL);

  gettimeofday(&tv, &tz);
 
  oldTime = (((tv.tv_sec*1000000) + tv.tv_usec));

  printf("Init %fs\n", (oldTime-curTime)/1000000000.0f);


  pthread_join(musicThread, NULL);
  pthread_join(demoThread, NULL);

  /* MikMod stop */
  Player_Stop();
  Player_Free(module);
  /* demo stuff */
  close3dEngine();
  /* libcaca stop */
  caca_free_bitmap(bitmap);
  caca_end();

  gettimeofday(&tv, &tz);
  curTime = ((tv.tv_sec*1000000) + tv.tv_usec)-baseTime;



  printf("Bye folks, duration %fmn\n", (curTime/1000000.0f)/60.0f);
  return 0;
}


    
void *musicFunction( void *ptr )
{
  Player_Start(module);
  while(musicPlay && Player_Active())
    {
    
      MikMod_Update();
      synchroCallback(module->sngpos, module->patpos);
      usleep(10000);
    }
  musicPlay = 0;
  return NULL;
}

void *demoFunction( void *ptr )
{
  x=0;y=0;z=0;

  while ((!caca_get_event(CACA_EVENT_KEY_PRESS)) && musicPlay) 
    {
      gettimeofday(&tv, &tz);
      oldTime = curTime;
      curTime = (((tv.tv_sec*1000000) + tv.tv_usec) - baseTime)/1000;

      timeline(curTime);
 
      /* Must be called AFTER caca_draw_bitmap() */
      // Noise((unsigned int*)screen, 10);

      if(curTime!=0)
	{
	  
	  frame = 1000/(curTime-oldTime);
	  sprintf(fpsdisp, "%d fps", frame);
	  	
	}
   
      frame++;	
    }


  musicPlay=0;
  return NULL;
}
void renderBuffer(void)
{
  caca_draw_bitmap(0, 0, caca_get_width() - 1, caca_get_height() - 1,bitmap, screen);
}
void renderScreen(void)
{
caca_set_color(CACA_COLOR_WHITE, CACA_COLOR_BLACK);
caca_putstr(0, caca_get_height()-1, fpsdisp);  
     caca_refresh();
}

void synchroCallback(int pos, int pat)
{
  if(pos==0 || pos==1)
    {
      curPart = 1;
      if(x>0)
	x--;
      if(MATCH_POSITION(0x24) || MATCH_POSITION(0x64))
	{
	  x+=10;
	}
    }
  else if(pos==2)
    {
      curPart = 2;
    }
  else if(pos==3)
    {
      curPart = 3;
    }


}

void timeline(int ticks)
{
  static float r=5;
  r-=0.1;
  memset(screen, 0, XSIZ*YSIZ*4);
  memset(scene->zbuffer, 100, XSIZ*YSIZ*4);
  if(curPart==0)
    {
   
    }
  
  switch(curPart)
    {
    case 1:
      displayLogo(ticks);
  break;
    case 2:
      displayPart2( ticks);
      break;
    case 3:
      displayPart3(ticks);
      break;
  

    default:
      break;
    }
  
}






/* DEMO BY ITSELF */

void displayLogo(int ticks)
{
  static float timeseek=0;
  static float timeseek2=0.5;
  sVector position;
 
  //caca_set_feature(CACA_DITHERING_FSTEIN);


 


  /* Logo */
  rotateObject(mandaObject,
	       x+cos(ticks/800.0f)*10,
	       180+ sin(ticks/1000.0f)*20,
	       180);
  moveObject(mandaObject,0,0,180);

  /* Particles Paths */
  timeseek+=0.01;
  if(timeseek>1.0)
    timeseek=0;
  getCurvePosition(mandaCurve, &position, timeseek,1);


  /* Particles Object positionning */
  rotateObject(particleGetObjectId(mandaParticles),
	       x+cos(ticks/800.0f)*10,
	       180+ sin(ticks/1000.0f)*20,
	       180);
  moveObject(particleGetObjectId(mandaParticles),0,0,180);
  
  updateParticles(mandaParticles, position.x, position.y, position.z,
		  ticks);

  rotateObject(particleGetObjectId(mandaParticles2),
	       x+cos(ticks/800.0f)*10,
	       180+ sin(ticks/1000.0f)*20,
	       180);
  moveObject(particleGetObjectId(mandaParticles2),0,0,180);
  

  timeseek2+=0.01;
  if(timeseek2>1.0)
    timeseek2=0;
  getCurvePosition(mandaCurve, &position, timeseek2,1);

  updateParticles(mandaParticles2, position.x, position.y, position.z,
		  ticks);
 
  /* Transform and draw */
  transformObjects();
  drawObjects(); 
  decayBlur((unsigned int*)screen, x);


  /* Particles are drawn separately */
  drawParticles(mandaParticles);
  drawParticles(mandaParticles2);
  renderBuffer();
  renderScreen();
  
}


void displayPart2(int ticks)
{

 
  static int i=0;


  static float t=0;
  //caca_set_feature(CACA_DITHERING_FSTEIN);

  for(i=0;i<PART2_COUNT;i++)
    {
      sVector position;
      getCurvePosition(part2Curve, &position, part2Time[i],1);
      part2Time[i]+=0.01;
      if(part2Time[i]>=1.0)
	part2Time[i] = 0;
  
      t+=0.2;

      updateParticles(part2Particles[i], position.x,position.y,position.z,ticks);
     
      moveObject(particleGetObjectId(part2Particles[i]),0,0,150);
      rotateObject(particleGetObjectId(part2Particles[i]),t,t/2.0f,180);


      transformObjects();
      drawParticles(part2Particles[i]);


    }
 
  drawString("Presents", 0,0, 1, 1,1,1);
  renderBuffer();
  printOutline(cabaret);
  renderScreen();

  /* 
     unsigned char str[1000];
     sprintf(str,"video/test%05d.ppm", dump);
     dumpPPM(str);

     dump++;
     printf("%d\n", dump);
  */
}



void displayPart3(int ticks)
{
  int i;
  static int nbr=0;
  static float t=0;
  unsigned char str[1024];
  t+=100;

  //caca_set_feature(CACA_DITHERING_FSTEIN);

  for(i=0;i<BALL_COUNT;i++)
    {
      scene->object[balls].vertex[i].x = (sin(i+(i+t/1000.0f))*0.3);
      scene->object[balls].vertex[i].y = (cos(i+(t/750.0f))*0.3);
      scene->object[balls].vertex[i].z = (sin(i+(t/800.0f))*0.2)+0.5;
      scene->object[balls].vertex[i].r = 8; 
    }
  
  
  moveObject(balls,0,0,0);
  rotateObject(balls,0,0,180);

  moveCamera(0.5,0.5,0);

  rotateObject(etoileObject,0,0,t/10.0f);
  moveObject(etoileObject,80,60,300);
  

  transformObjects();
    
  drawObjectFlatZbuffer(etoileObject);

  raytraceBalls((unsigned int*)screen, objectGet(balls),64,64);



  renderBuffer();
  renderScreen();


  /*
 sprintf(str,"video/test%05d.ppm", nbr);
      dumpPPM(str);

      nbr++;*/

}










void dumpPPM(char *filename)
{
  /*  sprintf(str,"video/test%05d.ppm", i);
      dumpPPM(str);

      i++;*/


  FILE *fp;
  int x,y;
  unsigned char *s=(unsigned char*)screen;
  fp=fopen(filename,"w+");
  

  fprintf(fp,"P3\n256 256\n256\n");
  for(y=0;y<YSIZ;y++)
    {
      for(x=0;x<XSIZ*4;x+=4)
	fprintf(fp,"%d %d %d ", 
		s[0+y*XSIZ*4], 
		s[1+x+y*XSIZ*4], 
		s[2+x+y*XSIZ*4]);

      fprintf(fp,"\n");
    }      


  fclose(fp);


}
