#include <demo.h>
#include <tunnel.h>
#include <3d.h>
#include <sprites.h>
sVector camera = { 0, 0, 0 };
sVector direction = { 0, 0, 0 };

Matrix cam_mat;            
unsigned int texture[65536];
point interpol[XSIZ+1][YSIZ+1];

float cam_rotx,cam_roty,cam_rotz;
float max=0,min=0;

#define SHADOW 100.0f
#define FOV 120
float speed=25;

void freeDirTunnel(void);
void freeDirPlane(void);
void normalize(sVector *v);
void crossProduct(sVector *v, sVector *v1, Matrix m1);
void matrixRotate(Matrix m, float rx, float ry, float rz);
void interpolation(void);


/* Heavily inspired by Chew's tutorial. Thanks hairy guy */


void tunnelCompute(unsigned int *tex, int ticks)
{
  int x, y;
  float r,g,b;
  int c, oy=0;
  unsigned int *s = (unsigned int*)screen;

  camera.z += speed;
  camera.x = 0;
  camera.y = 0;

  cam_rotx += .04;   
  cam_roty += .013;  
  cam_rotz += .09;   

  matrixRotate(cam_mat, cam_rotx, cam_roty, cam_rotz);


	
  freeDirTunnel();
  //freeDirPlane();
  interpolation();


  


  for(y=0;y<YSIZ;y++)
    {
      for(x=0;x<XSIZ;x++)
	{   
	  c = tex[((interpol[x][y].y<<8)+interpol[x][y].x)&0xffff];
	  r = (c & 0xFF0000)>>16;
	  g = (c & 0x00FF00)>>8;
	  b = c & 0x0000FF;
		
	  s[x+oy] =  MAKE_COLOR_CHAR((unsigned char)r,(unsigned char)g,(unsigned char)b);

	}
      oy+=XSIZ;
      
    }
}


void freeDirTunnel(void)
{
  int x, y;
  sVector d; 
  sVector inter; 

  float a, b, c;
  float delta;
  float t, t1, t2;

  static int rayon=256;

  c = (camera.x * camera.x) + (camera.y * camera.y) - (rayon*rayon);
 
  for(y=0;y<YSIZ+1;y+=8)
    for(x=0;x<XSIZ+1;x+=8)
      {
	d.x = ((float)(x-XSIZ/2))/FOV;
	d.y = ((float)(y-YSIZ/2))/FOV;
	d.z = 1;
	crossProduct(&d, &d, cam_mat);

	normalize(&d);

	a = (d.x*d.x) + (d.y*d.y);
	b = 2*(camera.x*d.x + camera.y*d.y);

	delta = (b*b) - 4*a*c;
	if(delta>0) {
	  delta = sqrt(delta);

	  t1 = (-b-delta)/(2*a);
	  if(t1>0) t=t1;

	  else
	    {
	      t2 = (-b+delta)/(2*a);
	      t = t2;
	    }

	  inter.x = camera.x + t*d.x;
	  inter.y = camera.y + t*d.y;
	  inter.z = camera.z + t*d.z;

	  interpol[x][y].x =
	    (unsigned short) (fabs( (256 * atan2(inter.y, inter.x) ) /M_PI) );
	  interpol[x][y].y =
	    (unsigned short)  fabs( inter.z );


	}
	else
	  {
	    interpol[x][y].x = 0;
	    interpol[x][y].y = 0;
	  }
      }
}


void freeDirPlane(void)
{
  int x, y;
  sVector d; 
  sVector inter; 

  float t;
  static int hauteur=256;

  for(y=0;y<YSIZ+1;y+=8)
    {
      for(x=0;x<XSIZ+1;x+=8)
	{
	  d.x = ((float)(x-160))/FOV;
	  d.y = ((float)(y-100))/FOV;
	  d.z = 1;
	  crossProduct(&d, &d, cam_mat);

	  normalize(&d);

	  if(d.y>0)
	    {
	      t = (hauteur - camera.y) / d.y;

	      inter.x = camera.x + t*d.x;
	      inter.y = camera.y + t*d.y;
	      inter.z = camera.z + t*d.z;

	      interpol[x][y].x = (unsigned short) fabs( inter.x );
	      interpol[x][y].y = (unsigned short) fabs( inter.z );
	    

	    }
	  else if(d.y<0)
	    {
	      t = (-hauteur - camera.y) / d.y;

	      inter.x = camera.x + t*d.x;
	      inter.y = camera.y + t*d.y;
	      inter.z = camera.z + t*d.z;

	      interpol[x][y].x = (unsigned short) fabs( inter.x );
	      interpol[x][y].y = (unsigned short) fabs( inter.z );
	
	      interpol[x][y].z = t;


	    }
	  else
	    {
	      // si il n'y a pas d'intersection
	      interpol[x][y].x = 0;
	      interpol[x][y].y = 0;
	      interpol[x][y].z = 0;
	    }
	   
	}
       
    }


}



void normalize(sVector *v)
{
  float l;

  l = sqrt((v->x*v->x) + (v->y*v->y) + (v->z*v->z));
  v->x = v->x/(l?l:1);
  v->y = v->y/(l?l:1);
  v->z = v->z/(l?l:1);
}

void crossProduct(sVector *v, sVector *v1, Matrix m1)
{
  int i;
  float temp[3] = { 0, 0, 0 };

  for(i=0;i<3;i++)
    {
      temp[i] += v1->x * m1[i][0];
      temp[i] += v1->y * m1[i][1];
      temp[i] += v1->z * m1[i][2];
    }

  v->x = temp[0];
  v->y = temp[1];
  v->z = temp[2];
}


void matrixRotate(Matrix m, float rx, float ry, float rz)
{
  float cosx, sinx;
  float cosy, siny;
  float cosz, sinz;

  cosx = (float) cos(rx);
  sinx = (float) sin(rx);
  cosy = (float) cos(ry);
  siny = (float) sin(ry);
  cosz = (float) cos(rz);
  sinz = (float) sin(rz);

  m[0][0] = cosy*cosz;
  m[0][1] = cosy*sinz;
  m[0][2] = -siny;

  m[1][0] = sinx*siny*cosz - cosx*sinz;
  m[1][1] = sinx*siny*sinz + cosx*cosz;
  m[1][2] = sinx*cosy;

  m[2][0] = cosx*siny*cosz + sinx*sinz;
  m[2][1] = cosx*siny*sinz - sinx*cosz;
  m[2][2] = cosx*cosy;
}

void interpolation(void)
{
  int _x, _y; 
  int x, y; 
  unsigned long  v; 
  unsigned long _v; 
  unsigned long dvdx; 
  unsigned long dvdy; 
  unsigned long dvdxdy;
  unsigned long v1, v2, v3, v4; 

  for(_y=0;_y<YSIZ;_y+=8)
    for(_x=0;_x<XSIZ;_x+=8)
      {
	v1 = interpol[_x][_y].x<<8; 
	v2 = interpol[_x+8][_y].x<<8; 
	v3 = interpol[_x][_y+8].x<<8; 
	v4 = interpol[_x+8][_y+8].x<<8;

	_v = v = v1;           

	dvdx   = (-v1+v2)>>3;
	dvdy   = (-v1+v3)>>3;
	dvdxdy = (v1-v2-v3+v4)>>6;

	for(y=_y;y<_y+8;y++) 
	  { 
	    for(x=_x;x<_x+8;x++) 
	      {
		v += dvdx;           
		interpol[x][y].x = v>>8; 
	      }
	    _v += dvdy;              
	    v = _v;                   
	    dvdx += dvdxdy;         
	  }

      }

  for(_y=0;_y<YSIZ;_y+=8) 
    for(_x=0;_x<XSIZ;_x+=8) 
      {
	v1 = interpol[_x][_y].y<<8; 
	v2 = interpol[_x+8][_y].y<<8;
	v3 = interpol[_x][_y+8].y<<8; 
	v4 = interpol[_x+8][_y+8].y<<8;

	_v = v = v1;                

	dvdx   = (-v1+v2)>>3;       
	dvdy   = (-v1+v3)>>3;       
	dvdxdy = (v1-v2-v3+v4)>>6;  

	for(y=_y;y<_y+8;y++) 
	  {      
	    for(x=_x;x<_x+8;x++)
	      { 
		v += dvdx;             
		interpol[x][y].y = v>>8;
	      }
	    _v   += dvdy;           
	    v     = _v;               
	    dvdx += dvdxdy;            
	  }
      }
}

